{-# LANGUAGE TypeFamilies #-}

module KNN (
  initialiseKNN
, featureTFIDFs
, knn
) where

import           Control.Arrow ((&&&), first)
import           Data.IntMap (IntMap)
import qualified Data.IntMap.Strict as IM
import qualified Data.IntSet as IS
import qualified Data.List as L
import           Data.Map (Map)
import           Data.Maybe (fromMaybe)
import           Data.Ord (comparing)

import           Predictor
import           Types
import           Utils

data KNN = KNN { samplesSeen :: Samples, fFreq :: FeatureFreq }

instance Predictor KNN where
  data PredictorCfg KNN = KNNConfig { depWeight :: Double }

  learn      = learnKNN
  predict    = predictKNN
  saveState  = saveStateKNN
  loadState  = loadStateKNN

type KNNCfg = PredictorCfg KNN
type KNNM   = PredictorM   KNN
type KNNT   = PredictorT   KNN

type LabelIM = IntMap
type Rank = (Int, Double)
type SWeight = (SampleID, Double)
type FWeights = IntMap Double
type Predictions = LabelIM [Rank]
type PredictionsN = (Int, Predictions)

defaultConf :: KNNCfg
defaultConf = KNNConfig {
  depWeight = 2.7
}

confActions :: [(String, KNNCfg -> String -> KNNCfg)]
confActions =
  [ ("depWeight", \ c v -> c {depWeight = read v})
  ]

initialiseKNN :: LearnData -> Map String String -> (KNN, KNNCfg)
initialiseKNN _ kvs =
  let conf = readConf confActions defaultConf kvs in
  (KNN {samplesSeen = [], fFreq = IM.empty}, conf)

learnKNN :: KNN -> Samples -> KNNM KNN
learnKNN (KNN sS fF) smpls = asksLData featureTable >>= \ fT ->
  return KNN {samplesSeen = sS ++ smpls, fFreq = featureFreq fT fF smpls}

predictKNN :: KNN -> Features -> KNNM Labels
predictKNN (KNN sS fF) ftrs = do
  (lT, fT) <- asksLData (labelTable &&& featureTable)
  nP <- asksLData nPredictions
  dW <- asksConfig depWeight

  return $ knn lT fT nP dW ftrs fWs $ zip sS $ repeat 1
  where fWs = featureTFIDFs (length sS) fF

saveStateKNN :: KNN -> FilePath -> KNNT IO ()
saveStateKNN _ _ = undefined

loadStateKNN :: KNN -> FilePath -> KNNT IO KNN
loadStateKNN _ _ = undefined


-- | Calculate importance of a feature from total number of samples and number
-- of appearances of the feature, using TF-IDF.
featureTFIDFs :: Int -> FeatureFreq -> FWeights
featureTFIDFs nS = IM.map (\ freq -> (nSL - log (fromIntegral freq)) ** 6)
  where nSL = log $ fromIntegral nS


-- | For each (sample, weight) in the list, determine the correlation (inverse
-- distance) of sample with features by multiplying sample weight with
-- sum of weights of common features.
-- For features without feature weight, use weight of 1.
-- Output (sample, correlation) list sorted by decreasing correlation.
correlate :: FeatureTable -> Features -> FWeights -> [SWeight] -> [SWeight]
correlate fT ftrs fWs = sort . foldr (uncurry insert) IM.empty where
  insert (SampleID s) w = IM.insertWith (+) s (w * score (SampleID s))
  score = sum . map weight . IS.toList . common
  weight = fromMaybe 1 . flip IM.lookup fWs
  common = IS.intersection ftrsS . IS.fromList . sampleFeatures fT
  sort = L.sortBy (flip $ comparing snd) . map (first SampleID) . IM.toList
  ftrsS = IS.fromList ftrs

-- | Predict labels from sorted (sample, correlation) list, stopping just
-- before the maximal number of predictions is exceeded.
predictLabels :: LabelTable -> Double -> Int -> [SWeight] -> Predictions
predictLabels lT lW maxPreds =
  snd . last . takeWhile p . scanl (addSample lT lW) (0, IM.empty) . zip [1 ..]
  where p (nPreds, _) = nPreds <= maxPreds

-- | Add sample and its labels (latter with label weight) to predictions.
addSample :: LabelTable -> Double -> PredictionsN -> (Int, SWeight) -> PredictionsN
addSample lT lW preds (k, (smpl, smplCorr)) =
  foldr (addLabel (k, lblCorr)) (addLabel (k, smplCorr) smplLbl preds) lbls
  where smplLbl = labelOfSample smpl
        lbls = sampleLabels lT smpl
        lblCorr = lW * smplCorr / log (fromIntegral (length lbls) + 1)

-- | Add label with rank to predictions.
addLabel :: Rank -> LabelID -> PredictionsN -> PredictionsN
addLabel rank lbl (nPreds, preds) = (nPreds', preds') where
  (oldR, preds') = IM.insertLookupWithKey (\ _ _ -> (rank :)) lbl [rank] preds
  -- | if for the label no ranks did exist before, increase nPreds
  nPreds' = maybe (nPreds + 1) (const nPreds) oldR

-- | Given a map from labels to their associated ranks, return labels sorted
-- by decreasing relevance.
sortLabels :: Predictions -> Labels
sortLabels = map fst . L.sortBy (flip $ comparing snd) . IM.toList . IM.map rank
  --map fst . L.sortBy (comparing snd) . IM.toList . IM.map rank
  where rank = sum . map (\ (k, corr) -> corr / log (fromIntegral k + 1))
  {-where rank xs = age (fst (head xs)) + sum (map snd xs)
        age k = 500000000 - 100 * fromIntegral k-}


-- | Run adaptive k-NN on weighted list of samples and list of features.
-- The resulting set of labels is sorted by decreasing relevance and contains
-- as many as possible, but less or equal than `maxPreds` labels.
-- The input samples are also output, but they are weighted differently than
-- their labels, because the latter are first weighted with `lblW`.
knn :: LabelTable -> FeatureTable -> Int -> Double -> Features ->
  FWeights -> [SWeight] -> Labels
knn lT fT maxPreds lblW ftrs fWs =
  sortLabels . predictLabels lT lblW maxPreds . correlate fT ftrs fWs
